"""
Constants based on the structure of the config and results storing system.
"""
# these are the ones you probably want to change if then
ROOT = "bcos"
BASE_EXPERIMENTS_DIRECTORY = "experiments"

# you probably don't want to change these
MODEL_FACTORY_MODULE = "model"
MODEL_FACTORY_VAR_NAME = "get_model"
CONFIGS_MODULE = "experiment_parameters"
CONFIGS_VAR_NAME = "CONFIGS"


CHECKPOINT_LAST_FILENAME = "last.ckpt"
